# Copyright 2017,2018 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from six.moves import configparser


class Opt(object):
    def __init__(self, opt_name, section='default',
                 opt_type='str', help='',
                 default=None, required=False):
        self.name = opt_name
        self.section = section
        self.opt_type = opt_type
        self.default = default
        self.required = required
        self.help = help


zvm_opts = [
    # logging options
    Opt('log_dir',
        section='logging',
        default='/var/log/zvmsdk/',
        help='''
Directory where log file to be put into.

SDK has a set of logs to help administrator to debug
and aduit actions performed through SDK. Edit this option
if you want to put logs into specified place.

Please ensure the service running on the consume which
consumes SDK has the authorization to write to the path.
    '''),
    Opt('log_level',
        section='logging',
        default='logging.INFO',
        help='''
Level of the log.

SDK utilize python logging package to help admin debug
or analyze issues. it's recommend to set this value
to logging.DEBUG to get more detailed logs and set it to
logging.INFO(default) in normal situation.

recommend values:
logging.ERROR: level above ERROR will be written to log file.
logging.WARNINS: level above WARNING(ERROR, WARNING)
                 will be written to log file.
logging.INFO: level above INFO(ERROR, WARNING, INFO)
              will be written to log file.
logging.DEBUG: All log level (ERROR, WARNING, INFO, DEBUG)
               will be written to log file.
    '''),
    # zvm options
    Opt('default_nic_vdev',
        section='zvm',
        default='1000',
        help='''
Virtual device number for default NIC address.

This value is the first NIC virtual device number,
each NIC needs 3 numbers for control/read/write, so by default
the first NIC's address is 1000, the second one is 1003 etc.

Possible values:
    An integer value in hex format, between 0 and 65536 (x'FFFF').
    It should not conflict with other device numbers in the z/VM guest's
    configuration, for example device numbers of the root or ephemeral or
    persistent disks.

Sample NIC definitions in the z/VM user directory:
    NICDEF 1000 TYPE QDIO LAN SYSTEM <vswitch1> MACID <macid1>
    NICDEF 1003 TYPE QDIO LAN SYSTEM <vswitch2> MACID <macid2>
        '''
        ),
    Opt('default_admin_userid',
        section='zvm',
        help='''
Default LOGONBY userid(s) for the cloud.

This is a set of z/VM userid(s) which are allowed to logon using the LOGONBY
keyword to the guests created by the z/VM SDK solution, compatible with
the LBYONLY keyword of the user directory statement. This value is only used
when a guest is created. If you change this value, existing guests' directory
entries are not automatically updated with the new value.
When an ESM is installed, this parameter only governs when the ESM
defers to CP's processing.

Usage note:
    The default is an empty string (''). When the string is empty, you can't
    log on to your instances using the 3270 protocol; When a
    non-empty string is provided, blank chars will be used as delimiter,
    you can use LOGONBY xxx command to log on the guest using the corresponding
    admin userid's password.

    For example, when you set this value to 'oper1 oper2 oper3 jones', it means
    you can use any one of 'oper1', 'oper2', 'oper3', 'jones' as an admin user.

    see the z/VM CP Planning and Administration for additional information.

Possible values:
    A maximum of 8 blank-delimited strings. Each non-blank string must be a
    valid z/VM userid.
    e.g  '' is a valid value.
         'oper1 oper2' is a valid value.
         'o1 o2 o3 o4 o5 o6 o7 o8 o9' is NOT a valid value.
    '''),
    # FIXME: remove this option when switch to smut
    Opt('user_default_password',
        section='zvm'),
    Opt('disk_pool',
        section='zvm',
        required=True,
        help='''
zVM disk pool and type for root/ephemeral disks.

The option is combined by 2 parts and use : as separator.

The first part is the type of disks in the disk pool.
The disks in one disk pool must in same type (ECKD or FBA).
Possible values of the disk pool type:
    A string, either ECKD or FBA.

The second part is the volume group name defined in your directory manager
on your z/VM system, which will be used for allocating disks for
new guest. A dollar sign ($) is not allowed in the name.

Sample disk_pool values:
    ECKD:diskpo1
    FBA:testpool
    '''),
    Opt('user_profile',
        section='zvm',
        required=True,
        help='''
PROFILE name to use when creating a z/VM guest.

When SDK deploys an guest on z/VM, it can include some
common statements from a PROFILE definition.
This PROFILE must already be included in your z/VM user directory.

Possible values:
    An 8 character name of a PROFILE that is already defined in the z/VM
    user directory.
    '''),
    Opt('user_root_vdev',
        section='zvm',
        default='0100',
        help='''
Virtual device number for root disk.

When SDK deploys an guest, it creates a root disk and potentially
several data disks. This value is the virtual device number of the root
disk.

Possible values:
    An integer value in hex format, between 0 and 65536 (x'FFFF').
    It should not conflict with other device numbers in the z/VM guest's
    configuration, for example device numbers of the NICs or ephemeral or
    persistent disks.

Sample root disk in user directory:
    MDISK 0100 <disktype> <start> <end> <volumelabel> <readwrite>
'''),
    Opt('namelist',
        section='zvm',
        help='''
The name of a list containing names of virtual servers to be queried. The list
which contains the userid list by default is named: VSMWORK1 NAMELIST, see
DMSSICNF COPY key: NameListFileIdAny. The list has to be accessible to the
SMAPI servers.

The length of namelist must no longer than 64.
'''),
    # image options
    Opt('temp_path',
        section='image',
        default='/tmp/zvmsdk/images/',
        help='''
Temp path to store image temp information.

During guest creation, image transferred from upper layer need to be
create some temp files, this option is used to tell SDK where
to store the temp files, make sure the process running SDK is able to
read write the directory.
    '''),
    # image options
    Opt('sdk_image_repository',
        section='image',
        default='/var/lib/zvmsdk/images',
        help='''
Directory to store sdk images.

SDK image repository to store the imported images and the staging images that
is in snapshotting. Once snapshot finished, the image will be removed to the
netboot directory accordingly. Two kinds of image repository looks like:
/var/lib/zvmsdk/images/netboot/<image_osversion>/<imagename>
/var/lib/zvmsdk/images/staging/<image_osversion>/<imagename>
    '''),
    # network options
    Opt('my_ip',
        section='network',
        required=True,
        help='''
IP address of the Linux machine which is running SDK on.

Some remote copy operations need to be performed during guest creation,
this option tell the SDK the host ip which can be used to perform copy
from and copy to operations.
    '''),
    # guest options
    Opt('temp_path',
        section='guest',
        default='/tmp/zvmsdk/guests/',
        help='''
Temp path to store guest temp information.

During guest creation, some temp data need to be stored in order to
be provided to guest later, this option is used to tell SDK where
to store the temp files, make sure the process running SDK is able to
read write the directory.
    '''),
    Opt('console_log_size',
        section='guest',
        default=100,
        opt_type='int',
        help='''
The maximum allowed console log size, in kilobytes.

Console logs might be transferred to sdk user, this option controls how
large each file can be. A smaller size may mean more calls will be needed
to transfer large consoles, which may not be desirable for performance reasons.
    '''),
    # monitor options
    Opt('cache_interval',
        section='monitor',
        default=300,
        opt_type='int',
        help='''
Cached monitor data update interval

This is used to prevent excessive effort spent retrieving the
monitor data by calling the SDK backend utilities. When this cache
is enabled, a inspect call will only call the SDK backend utilities
when the inspected guest's info does not exist in the cache or
when the cache data is expired. And when an cache update is needed,
all the existing guests' data will be retrieved in a single call to
the backend.

When this value is below or equal to zero, the cache
will be disabled and each inspect call will need to call the backend
utilities to get the inspected guest's monitor data.
        '''
        ),
    # wsgi options
    # this option is used when sending http request
    # to sdk wsgi, default to none so no token validation
    # will be used.
    Opt('auth',
        section='wsgi',
        default='none',
        opt_type='str',
        help='''
Whether auth will be used.

When sending http request from outside to running zvmsdk,
Client will be requested to input username/password in order
to authorize the call.
Set this to 'none' indicated no auth will be used and 'auth'
means username and password need to be specified.

Possible value:
'none': no auth will be required
'auth': need auth, currently pyjwt is used to return a token
        to caller if the username and password is correct.
''',
        ),
    Opt('token_validation_period',
        section='wsgi',
        default=3600,
        opt_type='int',
        help='''
How long the token is valid.

If a token auth is used, the token return to user will be
expired after the period passed. This ensure an user who
get this token will not be authorized all the time, a new
token need to be recreated after certain time period.
''',
        ),
    Opt('token_path',
        section='wsgi',
        default='/etc/zvmsdk/token.dat',
        opt_type='str',
        help='''
file path that contains admin-token to access sdk http server.

Admin-token in order to get a user-token from zvm sdk, and the user-token
will be used to validate request before user-token expire.
'''
        ),
    Opt('max_concurrent_deploy_capture',
        section='wsgi',
        default=20,
        opt_type='int',
        help='''
The max total number of concurrent deploy and capture requests allowed in a
single z/VM Cloud Connector process.

If more requests than this value are revieved concurrently, the z/VM Cloud
Connector would reject the requests and return error to avoid resource
exhaustion.
.
'''
        ),
    # Daemon server options
    Opt('bind_addr',
        section='sdkserver',
        default='127.0.0.1',
        opt_type='str',
        help='''
The IP address that the SDK server is listen on.

When the SDK server deamon starts, it will try to bind to
this address and port bind_port, and wait for the SDK client
connection to handle API request.
'''
        ),
    Opt('bind_port',
        section='sdkserver',
        opt_type='int',
        default=2000,
        help='''
The port that the SDK server is listen on.

This will work as a pair with bind_addr when the SDK server daemon
starts, more info can be found in that configuration description.
'''
        ),
    Opt('request_queue_size',
        section='sdkserver',
        opt_type='int',
        default=128,
        help='''
The size of request queue in SDK server.

SDK server maintains a queue to keep all the accepted but not handled requests,
and the SDK server workers fetch requests from this queue.
To some extend, this queue size decides the max socket opened in SDK server.
This value should be adjusted according to the system resource.
'''
        ),
    Opt('max_worker_count',
        section='sdkserver',
        opt_type='int',
        default=64,
        help='''
The maximum number of worker thread in SDK server to handle client requests.

These worker threads would work concurrently to handle requests from client.
This value should be adjusted according to the system resource and workload.
'''
        ),
    # database options
    Opt('dir',
        section='database',
        default='/var/lib/zvmsdk/databases/',
        opt_type='str',
        help='''
Directory to store database.

SDK databases are used to store a set of tables which contain the
information of network, volume, image, etc. This option is used to
tell SDK where to store the database files, make sure the process
running SDK is able to read write and execute the directory.
'''
        ),
    # tests options
    Opt('image_path',
        section='tests',
        opt_type='str',
        ),
    Opt('userid_prefix',
        section='tests',
        default='tst',
        opt_type='str',
        ),
    Opt('image_os_version',
        section='tests',
        opt_type='str',
        ),
    Opt('userid_list',
        section='tests',
        opt_type='str',
        ),
    Opt('ip_addr_list',
        section='tests',
        default='192.168.0.2 192.168.0.3 192.168.0.4 192.168.0.5 192.168.0.6',
        opt_type='str',
        ),
    Opt('mac_user_prefix',
        section='tests',
        default='02:00:00',
        opt_type='str',
        ),
    Opt('vswitch',
        section='tests',
        opt_type='str',
        ),
    Opt('broadcast_v4',
        section='tests'),
    Opt('gateway_v4',
        section='tests'),
    Opt('cidr',
        section='tests'),
    Opt('netmask_v4',
        section='tests'),
    ]


class ConfigOpts(object):

    def __init__(self):
        self.dicts = {}

    def get_config_dicts_default(self, opts):
        _dict = {}
        for opt in opts:
            sec = opt.section
            if _dict.get(sec) is None:
                _dict[sec] = {}
            _dict[sec][opt.name] = {'required': opt.required,
                                    'default': opt.default,
                                    'type': opt.opt_type,
                                    'help': opt.help}
        return _dict

    def register(self, opts):
        # Register the defined options and parse to dict
        self.dicts = self.get_config_dicts_default(opts)
        return self.clear_and_to_dict()

    def config(self):
        # Load config file and merge with default definitions
        # read config file
        override_dicts = self.read_config_file_to_dicts()
        # overwrite default value
        try:
            self.dicts = self.merge(self.dicts, override_dicts)
        except ImportError:
            pass
        # Check config value
        self._check_value(self.dicts)
        # Clear unused attributes of each option, and parse to our defined dict
        return self.clear_and_to_dict()

    def read_config_file_to_dicts(self):
        configs = {}
        read_file = self.find_config_file(project="zvmsdk")
        if read_file is None:
            raise ConfFileMissingError()
        else:
            cf = configparser.ConfigParser()
            cf.read(read_file)
            # read each section and option to dict
            secs = cf.sections()
            for sec in secs:
                configs[sec] = {}
                # get all options of the section in a list
                opts = cf.options(sec)
                for opt in opts:
                    val = cf.get(sec, opt)
                    configs[sec][opt] = val
            return configs

    def merge(self, defaults, override):
        # merge the defaults and overridden
        # The overridden options would only have 'default' set in the
        # resulted dicts
        r = {}
        for k, v in defaults.items():
            if k in override:
                if isinstance(v, dict) and isinstance(override[k], dict):
                    r[k] = self.merge(v, override[k])
                elif isinstance(v, dict):
                    if override[k] is not None:
                        v['default'] = override[k]
                    r[k] = v
                else:
                    r[k] = override[k]
            else:
                r[k] = v
        return r

    def clear_and_to_dict(self):
        # This function would clear the dict to remove the unused keys
        # ('required', 'default', 'type', 'help'), set the opt value to
        # the final value merged in 'default'.
        # And then, convert the python dict to our defined Dict object
        clear_dict = {}
        pydict = self.dicts
        for k1, v1 in pydict.items():
            r_con = {}
            for k2, v2 in v1.items():
                r_con[k2] = v2['default']
            clear_dict[k1] = r_con

        return self.toDict(clear_dict)

    def _check_value(self, conf):
        for k1, v1 in conf.items():
            for k2, v2 in v1.items():
                # Check required options
                if v2['required'] and (v2['default'] is None):
                    raise RequiredOptMissingError(k1, k2)
                # Convert type
                if v2['type'] == 'int':
                    v2['default'] = int(v2['default'])
                # Check format
                if (k2 == "disk_pool") and (v2['default'] is not None):
                    self._check_zvm_disk_pool(v2['default'])

    def _check_zvm_disk_pool(self, value):
        disks = value.split(':')
        if (len(disks) != 2) or (disks[0].upper() not in ['FBA', 'ECKD']) or (
            disks[1] == ''):
            raise OptFormatError("zvm", "disk_pool", value)

    def toDict(self, d):
        D = Dict()
        for k, v in d.items():
            D[k] = self.toDict(v) if isinstance(v, dict) else v
        return D

    def _fixpath(self, p):
        """Apply tilde expansion and absolutization to a path."""
        return os.path.abspath(os.path.expanduser(p))

    def _get_config_dirs(self):
        """Return a list of directories where config files may be located.

        following directories are returned::

          ./
          ../etc
          ~/
          /etc/zvmsdk/
        """
        _cwd = os.path.split(os.path.abspath(__file__))[0]
        _pdir = os.path.split(_cwd)[0]
        _etcdir = ''.join((_pdir, '/', 'etc/'))
        cfg_dirs = [
            self._fixpath(_cwd),
            self._fixpath('/etc/zvmsdk/'),
            self._fixpath('/etc/'),
            self._fixpath('~'),
            self._fixpath(_etcdir),
        ]
        return [x for x in cfg_dirs if x]

    def _search_dirs(self, dirs, basename, extension=""):
        """Search a list of directories for a given filename or directory name.

        Iterator over the supplied directories, returning the first file
        found with the supplied name and extension.

        :param dirs: a list of directories
        :param basename: the filename
        :param extension: the file extension, for example '.conf'
        :returns: the path to a matching file, or None
        """
        for d in dirs:
            path = os.path.join(d, '%s%s' % (basename, extension))
            if os.path.exists(path):
                return path

        return None

    def find_config_file(self, project=None, extension='.conf'):
        """Return the config file.

        :param project: "zvmsdk"
        :param extension: the type of the config file

        """
        cfg_dirs = self._get_config_dirs()
        config_files = self._search_dirs(cfg_dirs, project, extension)

        return config_files


class Dict(dict):
    '''
    Simple dict but support access as x.y style.
    '''
    def __init__(self, names=(), values=(), **kw):
        super(Dict, self).__init__(**kw)
        for k, v in zip(names, values):
            self[k] = v

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'CONF' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        self[key] = value


class RequiredOptMissingError(Exception):
    """Raised if an option is required but no value is supplied by the user."""

    def __init__(self, grp_name, opt_name):
        self.grp_name = grp_name
        self.opt_name = opt_name

    def __str__(self):
        return "value required for option %s - %s" % (self.grp_name,
                                                      self.opt_name)


class ConfFileMissingError(Exception):
    """Raised if the configuration file zvmsdk.conf cann't be found."""

    def __init__(self):
        message = "zvmsdk.conf is not found."
        super(ConfFileMissingError, self).__init__(message)


class OptFormatError(Exception):
    """Raised if an option is required but no value is supplied by the user."""

    def __init__(self, grp_name, opt_name, value):
        self.grp_name = grp_name
        self.opt_name = opt_name
        self.value = value

    def __str__(self):
        return "value %s for option  %s - %s is invalid" % (self.value,
                                                            self.grp_name,
                                                            self.opt_name)


CONFOPTS = ConfigOpts()
CONF = CONFOPTS.register(zvm_opts)


def load_config():
    global CONF
    CONF = CONFOPTS.config()



# Copyright 2017,2018 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

HYPERVISOR_TYPE = 'zvm'
ARCHITECTURE = 's390x'
ALLOWED_VM_TYPE = 'zLinux'
PROV_METHOD = 'netboot'
ZVM_USER_DEFAULT_PRIVILEGE = 'G'
CONFIG_DRIVE_FORMAT = 'tgz'
DEFAULT_EPH_DISK_FMT = 'ext3'
DISK_FUNC_NAME = 'setupDisk'

RINV_HOST_KEYWORDS = {
    "zvm_host": "z/VM Host:",
    "zhcp": "zHCP:",
    "cec_vendor": "CEC Vendor:",
    "cec_model": "CEC Model:",
    "hypervisor_os": "Hypervisor OS:",
    "hypervisor_name": "Hypervisor Name:",
    "architecture": "Architecture:",
    "lpar_cpu_total": "LPAR CPU Total:",
    "lpar_cpu_used": "LPAR CPU Used:",
    "lpar_memory_total": "LPAR Memory Total:",
    "lpar_memory_used": "LPAR Memory Used:",
    "lpar_memory_offline": "LPAR Memory Offline:",
    "ipl_time": "IPL Time:",
    }

DISKPOOL_KEYWORDS = {
    "disk_total": "Total:",
    "disk_used": "Used:",
    "disk_available": "Free:",
    }

SET_VSWITCH_KEYWORDS = ["grant_userid", "user_vlan_id",
                        "revoke_userid", "real_device_address",
                        "port_name", "controller_name",
                        "connection_value", "queue_memory_limit",
                        "routing_value", "port_type", "persist",
                        "gvrp_value", "mac_id", "uplink",
                        "nic_userid", "nic_vdev",
                        "lacp", "interval", "group_rdev",
                        "iptimeout", "port_isolation", "promiscuous",
                        "MAC_protect", "VLAN_counters"]

DEV_STATUS = {'0': 'Device is not active.',
              '1': 'Device is active.',
              '2': 'Device is a backup device'}

DEV_ERROR = {'0': 'No error.',
             '1': 'Port name conflict.',
             '2': 'No layer 2 support.',
             '3': 'Real device does not exist.',
             '4': 'Real device is attached elsewhere.',
             '5': 'Real device is not compatible type.',
             '6': 'Initialization error.',
             '7': 'Stalled OSA.',
             '8': 'Stalled controller.',
             '9': 'Controller connection severed.',
             '10': 'Primary or secondary routing conflict.',
             '11': 'Device is offline.',
             '12': 'Device was detached.',
             '13': 'IP/Ethernet type mismatch.',
             '14': 'Insufficient memory in controller '
                   'virtual machine.',
             '15': 'TCP/IP configuration conflict.',
             '16': 'No link aggregation support.',
             '17': 'OSA-E attribute mismatch.',
             '18': 'Reserved for future use.',
             '19': 'OSA-E is not ready.',
             '20': 'Reserved for future use.',
             '21': 'Attempting restart for device.',
             '22': 'Exclusive user error.',
             '23': 'Device state is invalid.',
             '24': 'Port number is invalid for device.',
             '25': 'No OSA connection isolation.',
             '26': 'EQID mismatch.',
             '27': 'Incompatible controller.',
             '28': 'BACKUP detached.',
             '29': 'BACKUP not ready.',
             '30': 'BACKUP attempting restart.',
             '31': 'EQID mismatch.',
             '32': 'No HiperSockets bridge support.',
             '33': 'HiperSockets bridge error.'}

SWITCH_STATUS = {'1': 'Virtual switch defined.',
                 '2': 'Controller not available.',
                 '3': 'Operator intervention required.',
                 '4': 'Disconnected.',
                 '5': 'Virtual devices attached to controller. '
                      'Normally a transient state.',
                 '6': 'OSA initialization in progress. '
                      'Normally a transient state.',
                 '7': 'OSA device not ready',
                 '8': 'OSA device ready.',
                 '9': 'OSA devices being detached. '
                      'Normally a transient state.',
                 '10': 'Virtual switch delete pending. '
                       'Normally a transient state.',
                 '11': 'Virtual switch failover recovering. '
                       'Normally a transient state.',
                 '12': 'Autorestart in progress. '
                       'Normally a transient state.'}

ZVM_VOLUMES_FILE = 'zvm_volumes'
ZVM_VOLUME_STATUS = ['free', 'in-use']
VOLUME_MULTI_PASS = 'MULTI'

POWER_STATE_ON = u'on'
POWER_STATE_OFF = u'off'

DATABASE_VOLUME = 'sdk_volume.sqlite'
DATABASE_NETWORK = 'sdk_network.sqlite'
DATABASE_GUEST = 'sdk_guest.sqlite'
DATABASE_IMAGE = 'sdk_image.sqlite'

IMAGE_TYPE = {
    'DEPLOY': 'netboot',
    'CAPTURE': 'staging'}

SDK_DATA_PATH = '/var/lib/zvmsdk/'



# Power functions for Systems Management Ultra Thin Layer
#
# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from smutLayer import generalUtils
from smutLayer import msgs
from smutLayer.vmUtils import execCmdThruIUCV, invokeSMCLI
from smutLayer.vmUtils import isLoggedOn
from smutLayer.vmUtils import waitForOSState, waitForVMState

modId = 'PVM'
vmOSUpStates = ['on', 'up']
vmOSUpDownStates = ['down', 'off', 'on', 'up']
version = "1.0.0"

"""
List of subfunction handlers.
Each subfunction contains a list that has:
  Readable name of the routine that handles the subfunction,
  Code for the function call.
"""
subfuncHandler = {
    'HELP': ['help', lambda rh: help(rh)],
    'ISREACHABLE': ['checkIsReachable',
        lambda rh: checkIsReachable(rh)],
    'OFF': ['deactivate', lambda rh: deactivate(rh)],
    'ON': ['activate', lambda rh: activate(rh)],
    'PAUSE': ['pause', lambda rh: pause(rh)],
    'REBOOT': ['reboot', lambda rh: reboot(rh)],
    'RESET': ['reset', lambda rh: reset(rh)],
    'SOFTOFF': ['softDeactivate', lambda rh: softDeactivate(rh)],
    'STATUS': ['getStatus', lambda rh: getStatus(rh)],
    'UNPAUSE': ['unpause', lambda rh: unpause(rh)],
    'VERSION': ['getVersion', lambda rh: getVersion(rh)],
    'WAIT': ['wait', lambda rh: wait(rh)],
    }

"""
List of positional operands based on subfunction.
Each subfunction contains a list which has a dictionary with the following
information for the positional operands:
  - Human readable name of the operand,
  - Property in the parms dictionary to hold the value,
  - Is it required (True) or optional (False),
  - Type of data (1: int, 2: string).
"""
posOpsList = {}

"""
List of additional operands/options supported by the various subfunctions.
The dictionary followng the subfunction name uses the keyword from the
command as a key.  Each keyword has a dictionary that lists:
  - the related parms item that stores the value,
  - how many values follow the keyword, and
  - the type of data for those values (1: int, 2: string)
For example, the 'WAIT' subfunction has a 'poll' operand that takes
one additional operand (time in seconds) which is an int.
While the 'showparms' operand is just the keyword and has no additional
value portion.
"""
keyOpsList = {
    'HELP': {},
    'ISREACHABLE': {
        '--showparms': ['showParms', 0, 0]},
    'OFF': {
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--wait': ['wait', 0, 0]},
    'ON': {
        '--state': ['desiredState', 1, 2],
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--wait': ['wait', 0, 0]},
    'PAUSE': {'--showparms': ['showParms', 0, 0]},
    'REBOOT': {
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--wait': ['wait', 0, 0]},
    'RESET': {
        '--state': ['desiredState', 1, 2],
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--wait': ['wait', 0, 0]},
    'SOFTOFF': {
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--wait': ['wait', 0, 0]},
    'STATUS': {
        '--showparms': ['showParms', 0, 0]
        },
    'UNPAUSE': {
        '--showparms': ['showParms', 0, 0]},
    'VERSION': {},
    'WAIT': {
        '--maxwait': ['maxWait', 1, 1],
        '--poll': ['poll', 1, 1],
        '--showparms': ['showParms', 0, 0],
        '--state': ['desiredState', 1, 2]},
    }


def activate(rh):
    """
    Activate a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'ON'
          userid      - userid of the virtual machine
          parms['desiredState']   - Desired state. Optional,
                                    unless 'maxQueries' is specified.
          parms['maxQueries']     - Maximum number of queries to issue.
                                    Optional.
          parms['maxWait']        - Maximum time to wait in seconds. Optional,
                                    unless 'maxQueries' is specified.
          parms['poll']           - Polling interval in seconds. Optional,
                                    unless 'maxQueries' is specified.

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """
    rh.printSysLog("Enter powerVM.activate, userid: " + rh.userid)

    parms = ["-T", rh.userid]
    smcliResults = invokeSMCLI(rh, "Image_Activate", parms)
    if smcliResults['overallRC'] == 0:
        pass
    elif (smcliResults['overallRC'] == 8 and
        smcliResults['rc'] == 200 and smcliResults['rs'] == 8):
        pass    # All good.  No need to change the ReqHandle results.
    else:
        # SMAPI API failed.
        rh.printLn("ES", smcliResults['response'])
        rh.updateResults(smcliResults)    # Use results from invokeSMCLI

    if rh.results['overallRC'] == 0 and 'maxQueries' in rh.parms:
        # Wait for the system to be in the desired state of:
        # OS is 'up' and reachable or VM is 'on'.
        if rh.parms['desiredState'] == 'up':
            results = waitForOSState(
                rh,
                rh.userid,
                rh.parms['desiredState'],
                maxQueries=rh.parms['maxQueries'],
                sleepSecs=rh.parms['poll'])
        else:
            results = waitForVMState(
                rh,
                rh.userid,
                rh.parms['desiredState'],
                maxQueries=rh.parms['maxQueries'],
                sleepSecs=rh.parms['poll'])

        if results['overallRC'] == 0:
            rh.printLn("N", "%s: %s" %
                (rh.userid, rh.parms['desiredState']))
        else:
            rh.updateResults(results)

    rh.printSysLog("Exit powerVM.activate, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def checkIsReachable(rh):
    """
    Check if a virtual machine is reachable.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'ISREACHABLE'
          userid      - userid of the virtual machine

    Output:
       Request Handle updated with the results.
       overallRC - 0: determined the status, non-zero: some weird failure
                                             while trying to execute a command
                                             on the guest via IUCV
       rc - RC returned from execCmdThruIUCV
       rs - 0: not reachable, 1: reachable
    """

    rh.printSysLog("Enter powerVM.checkIsReachable, userid: " +
        rh.userid)

    strCmd = "echo 'ping'"
    results = execCmdThruIUCV(rh, rh.userid, strCmd)

    if results['overallRC'] == 0:
        rh.printLn("N", rh.userid + ": reachable")
        reachable = 1
    else:
        # A failure from execCmdThruIUCV is acceptable way of determining
        # that the system is unreachable.  We won't pass along the
        # error message.
        rh.printLn("N", rh.userid + ": unreachable")
        reachable = 0

    rh.updateResults({"rs": reachable})
    rh.printSysLog("Exit powerVM.checkIsReachable, rc: 0")
    return 0


def deactivate(rh):
    """
    Deactivate a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'OFF'
          userid      - userid of the virtual machine
          parms['maxQueries']     - Maximum number of queries to issue.
                                    Optional.
          parms['maxWait']        - Maximum time to wait in seconds. Optional,
                                    unless 'maxQueries' is specified.
          parms['poll']           - Polling interval in seconds. Optional,
                                    unless 'maxQueries' is specified.

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.deactivate, userid: " +
        rh.userid)

    parms = ["-T", rh.userid, "-f", "IMMED"]
    results = invokeSMCLI(rh, "Image_Deactivate", parms)
    if results['overallRC'] == 0:
        pass
    elif (results['overallRC'] == 8 and results['rc'] == 200 and
        (results['rs'] == 12 or results['rs'] == 16)):
        # Tolerable error.  Machine is already in or going into the state
        # we want it to enter.
        rh.printLn("N", rh.userid + ": off")
        rh.updateResults({}, reset=1)
    else:
        # SMAPI API failed.
        rh.printLn("ES", results['response'])
        rh.updateResults(results)    # Use results from invokeSMCLI

    if results['overallRC'] == 0 and 'maxQueries' in rh.parms:
        results = waitForVMState(
            rh,
            rh.userid,
            'off',
            maxQueries=rh.parms['maxQueries'],
            sleepSecs=rh.parms['poll'])
        if results['overallRC'] == 0:
            rh.printLn("N", rh.userid + ": off")
        else:
            rh.updateResults(results)

    rh.printSysLog("Exit powerVM.deactivate, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def doIt(rh):
    """
    Perform the requested function by invoking the subfunction handler.

    Input:
       Request Handle

    Output:
       Request Handle updated with parsed input.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.doIt")

    # Show the invocation parameters, if requested.
    if 'showParms' in rh.parms and rh.parms['showParms'] is True:
        rh.printLn("N", "Invocation parameters: ")
        rh.printLn("N", "  Routine: powerVM." +
            str(subfuncHandler[rh.subfunction][0]) + "(reqHandle)")
        rh.printLn("N", "  function: " + rh.function)
        rh.printLn("N", "  userid: " + rh.userid)
        rh.printLn("N", "  subfunction: " + rh.subfunction)
        rh.printLn("N", "  parms{}: ")
        for key in rh.parms:
            if key != 'showParms':
                rh.printLn("N", "    " + key + ": " +
                    str(rh.parms[key]))
        rh.printLn("N", " ")

    # Call the subfunction handler
    subfuncHandler[rh.subfunction][1](rh)

    rh.printSysLog("Exit powerVM.doIt, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def getStatus(rh):
    """
    Get the power (logon/off) status of a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'STATUS'
          userid      - userid of the virtual machine

    Output:
       Request Handle updated with the results.
       results['overallRC'] - 0: ok, non-zero: error
       if ok:
          results['rc'] - 0: for both on and off cases
          results['rs'] - 0: powered on
          results['rs'] - 1: powered off
    """

    rh.printSysLog("Enter powerVM.getStatus, userid: " +
        rh.userid)

    results = isLoggedOn(rh, rh.userid)
    if results['overallRC'] != 0:
        # Unexpected error
        pass
    elif results['rs'] == 0:
        rh.printLn("N", rh.userid + ": on")
    else:
        rh.printLn("N", rh.userid + ": off")

    rh.updateResults(results)

    rh.printSysLog("Exit powerVM.getStatus, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def getVersion(rh):
    """
    Get the version of this function.

    Input:
       Request Handle

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printLn("N", "Version: " + version)
    return 0


def help(rh):
    """
    Produce help output specifically for PowerVM functions.

    Input:
       Request Handle

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    showInvLines(rh)
    showOperandLines(rh)
    return 0


def parseCmdline(rh):
    """
    Parse the request command input.

    Input:
       Request Handle

    Output:
       Request Handle updated with parsed input.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.parseCmdline")

    if rh.totalParms >= 2:
        rh.userid = rh.request[1].upper()
    else:
        # Userid is missing.
        msg = msgs.msg['0010'][1] % modId
        rh.printLn("ES", msg)
        rh.updateResults(msgs.msg['0010'][0])
        rh.printSysLog("Exit powerVM.parseCmdLine, rc: " +
            rh.results['overallRC'])
        return rh.results['overallRC']

    if rh.totalParms == 2:
        rh.subfunction = rh.userid
        rh.userid = ''

    if rh.totalParms >= 3:
        rh.subfunction = rh.request[2].upper()

    # Verify the subfunction is valid.
    if rh.subfunction not in subfuncHandler:
        # Subfunction is missing.
        subList = ', '.join(sorted(subfuncHandler.keys()))
        msg = msgs.msg['0011'][1] % (modId, subList)
        rh.printLn("ES", msg)
        rh.updateResults(msgs.msg['0011'][0])

    # Parse the rest of the command line.
    if rh.results['overallRC'] == 0:
        rh.argPos = 3               # Begin Parsing at 4th operand
        generalUtils.parseCmdline(rh, posOpsList, keyOpsList)

    waiting = 0
    if rh.results['overallRC'] == 0:
        if rh.subfunction == 'WAIT':
            waiting = 1
            if rh.parms['desiredState'] not in vmOSUpDownStates:
                # Desired state is not: down, off, on or up.
                msg = msgs.msg['0013'][1] % (modId,
                    rh.parms['desiredState'], ", ".join(vmOSUpDownStates))
                rh.printLn("ES", msg)
                rh.updateResults(msgs.msg['0013'][0])

    if (rh.results['overallRC'] == 0 and 'wait' in rh.parms):
        waiting = 1
        if 'desiredState' not in rh.parms:
            if rh.subfunction in ['ON', 'RESET', 'REBOOT']:
                rh.parms['desiredState'] = 'up'
            else:
                # OFF and SOFTOFF default to 'off'.
                rh.parms['desiredState'] = 'off'

    if rh.results['overallRC'] == 0 and waiting == 1:
        if rh.subfunction == 'ON' or rh.subfunction == 'RESET':
            if ('desiredState' not in rh.parms or
                  rh.parms['desiredState'] not in vmOSUpStates):
                # Desired state is not: on or up.
                msg = msgs.msg['0013'][1] % (modId,
                    rh.parms['desiredState'], ", ".join(vmOSUpStates))
                rh.printLn("ES", msg)
                rh.updateResults(msgs.msg['0013'][0])

        if rh.results['overallRC'] == 0:
            if 'maxWait' not in rh.parms:
                rh.parms['maxWait'] = 300
            if 'poll' not in rh.parms:
                rh.parms['poll'] = 15
            rh.parms['maxQueries'] = (rh.parms['maxWait'] +
                rh.parms['poll'] - 1) / rh.parms['poll']
            # If we had to do some rounding, give a warning
            # out to the command line user that the wait
            # won't be what they expected.
            if rh.parms['maxWait'] % rh.parms['poll'] != 0:
                msg = msgs.msg['0017'][1] % (modId,
                    rh.parms['maxWait'], rh.parms['poll'],
                    rh.parms['maxQueries'] * rh.parms['poll'],
                    rh.parms['maxQueries'])
                rh.printLn("W", msg)

    rh.printSysLog("Exit powerVM.parseCmdLine, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def pause(rh):
    """
    Pause a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'PAUSE'
          userid      - userid of the virtual machine

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.pause, userid: " + rh.userid)

    parms = ["-T", rh.userid, "-k", "PAUSE=YES"]
    results = invokeSMCLI(rh, "Image_Pause", parms)
    if results['overallRC'] != 0:
        # SMAPI API failed.
        rh.printLn("ES", results['response'])
        rh.updateResults(results)    # Use results from invokeSMCLI

    rh.printSysLog("Exit powerVM.pause, rc: " + str(rh.results['overallRC']))
    return rh.results['overallRC']


def reboot(rh):
    """
    Reboot a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'REBOOT'
          userid      - userid of the virtual machine
          parms['desiredState']   - Desired state. Optional, unless
                                    'maxQueries' is specified.
          parms['maxQueries']     - Maximum number of queries to issue.
                                    Optional.
          parms['maxWait']        - Maximum time to wait in seconds. Optional,
                                    unless 'maxQueries' is specified.
          parms['poll']           - Polling interval in seconds. Optional,
                                    unless 'maxQueries' is specified.

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.reboot, userid: " + rh.userid)

    strCmd = "shutdown -r now"
    results = execCmdThruIUCV(rh, rh.userid, strCmd)
    if results['overallRC'] != 0:
        # Command failed to execute using IUCV.
        rh.printLn("ES", results['response'])
        rh.updateResults(results)

    if rh.results['overallRC'] == 0:
        # Wait for the OS to go down
        results = waitForOSState(rh, rh.userid, "down",
            maxQueries=30, sleepSecs=10)
        if results['overallRC'] == 0:
            rh.printLn("N", rh.userid + ": down (interim state)")

    if rh.results['overallRC'] == 0 and 'maxQueries' in rh.parms:
        results = waitForOSState(rh,
                                  rh.userid,
                                  'up',
                                  maxQueries=rh.parms['maxQueries'],
                                  sleepSecs=rh.parms['poll'])
        if results['overallRC'] == 0:
            rh.printLn("N", rh.userid + ": up")
        else:
            rh.updateResults(results)

    rh.printSysLog("Exit powerVM.reboot, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def reset(rh):
    """
    Reset a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'RESET'
          userid      - userid of the virtual machine
          parms['maxQueries']     - Maximum number of queries to issue.
                                    Optional.
          parms['maxWait']        - Maximum time to wait in seconds. Optional,
                                    unless 'maxQueries' is specified.
          parms['poll']           - Polling interval in seconds. Optional,
                                    unless 'maxQueries' is specified.

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.reset, userid: " + rh.userid)

    # Log off the user
    parms = ["-T", rh.userid]
    results = invokeSMCLI(rh, "Image_Deactivate", parms)
    if results['overallRC'] != 0:
        if results['rc'] == 200 and results['rs'] == 12:
            # Tolerated error.  Machine is already in the desired state.
            results['overallRC'] = 0
            results['rc'] = 0
            results['rs'] = 0
        else:
            # SMAPI API failed.
            rh.printLn("ES", results['response'])
            rh.updateResults(results)    # Use results from invokeSMCLI

    # Wait for the logoff to complete
    if results['overallRC'] == 0:
        results = waitForVMState(rh, rh.userid, "off",
            maxQueries=30, sleepSecs=10)

    # Log the user back on
    if results['overallRC'] == 0:
        parms = ["-T", rh.userid]
        results = invokeSMCLI(rh, "Image_Activate", parms)
        if results['overallRC'] != 0:
            # SMAPI API failed.
            rh.printLn("ES", results['response'])
            rh.updateResults(results)    # Use results from invokeSMCLI

    if results['overallRC'] == 0 and 'maxQueries' in rh.parms:
        if rh.parms['desiredState'] == 'up':
            results = waitForOSState(
                rh,
                rh.userid,
                rh.parms['desiredState'],
                maxQueries=rh.parms['maxQueries'],
                sleepSecs=rh.parms['poll'])
        else:
            results = waitForVMState(
                rh,
                rh.userid,
                rh.parms['desiredState'],
                maxQueries=rh.parms['maxQueries'],
                sleepSecs=rh.parms['poll'])
        if results['overallRC'] == 0:
            rh.printLn("N", rh.userid + ": " +
                       rh.parms['desiredState'])
        else:
            rh.updateResults(results)

    rh.printSysLog("Exit powerVM.reset, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def showInvLines(rh):
    """
    Produce help output related to command synopsis

    Input:
       Request Handle
    """

    if rh.subfunction != '':
        rh.printLn("N", "Usage:")
    rh.printLn("N", "  python " + rh.cmdName +
        " PowerVM <userid>")
    rh.printLn("N", "                    [isreachable | pause | " +
        "status | unpause]")
    rh.printLn("N", "  python " + rh.cmdName +
        " PowerVM <userid>")
    rh.printLn("N", "                    [on | reset] --wait --state " +
        "[on | up] --maxwait <secs>")
    rh.printLn("N", "                    --poll <secs>")
    rh.printLn("N", "  python " + rh.cmdName +
        " PowerVM <userid>")
    rh.printLn("N", "                    [off | reboot | softoff] " +
        "--maxwait <secs> --poll <secs>")
    rh.printLn("N", "  python " + rh.cmdName + " PowerVM " +
        "<userid> wait")
    rh.printLn("N", "                    --state [down | on | off | up] " +
        "--maxwait <secs>")
    rh.printLn("N", "                    --poll <secs>")
    rh.printLn("N", "  python " + rh.cmdName + " PowerVM help")
    rh.printLn("N", "  python " + rh.cmdName + " PowerVM version")
    return


def showOperandLines(rh):
    """
    Produce help output related to operands.

    Input:
       Request Handle
    """

    if rh.function == 'HELP':
        rh.printLn("N", "  For the PowerVM function:")
    else:
        rh.printLn("N", "Sub-Functions(s):")
    rh.printLn("N", "      help        - Displays this help " +
        "information.")
    rh.printLn("N", "      isreachable - Determine whether the " +
        "virtual OS in a virtual machine")
    rh.printLn("N", "                    is reachable")
    rh.printLn("N", "      on          - Log on the virtual machine")
    rh.printLn("N", "      off         - Log off the virtual machine")
    rh.printLn("N", "      pause       - Pause a virtual machine")
    rh.printLn("N", "      reboot      - Issue a shutdown command to " +
        "reboot the OS in a virtual")
    rh.printLn("N", "                    machine")
    rh.printLn("N", "      reset       - Power a virtual machine off " +
        "and then back on")
    rh.printLn("N", "      softoff     - Issue a shutdown command to " +
        "shutdown the OS in a virtual")
    rh.printLn("N", "                    machine and then log the " +
        "virtual machine off z/VM.")
    rh.printLn("N", "      status      - show the log on/off status " +
        "of the virtual machine")
    rh.printLn("N", "      unpause     - Unpause a virtual machine")
    rh.printLn("N", "      wait        - Wait for the virtual machine " +
        "to go into the specified")
    rh.printLn("N", "                    state of either:")
    rh.printLn("N", "                       down: virtual machine's " +
        "OS is not reachable with IUCV")
    rh.printLn("N", "                       off: virtual machine is " +
        "logged off")
    rh.printLn("N", "                       on: virtual machine is " +
        "logged on")
    rh.printLn("N", "                       up: virtual machine's OS " +
        "is reachable with IUCV")
    rh.printLn("N", "      version     - show the version of the " +
        "power function")
    if rh.subfunction != '':
        rh.printLn("N", "Operand(s):")
    rh.printLn("N", "      <userid>    - Userid of the target " +
        "virtual machine")
    rh.printLn("N", "      --maxwait <secs> - " +
        "Maximum time in seconds to wait")
    rh.printLn("N", "      --poll <secs> - " +
        "Seconds to wait between polling")
    rh.printLn("N", "      --state [down | off | on | up] - " +
        "Desired state for virtual machine")
    rh.printLn("N", "                    (on or off) or for the operating " +
        "system (down or up).")
    rh.printLn("N", "      --wait      - wait for the machine to go into " +
        "the desired state.")

    return


def softDeactivate(rh):
    """
    Deactivate a virtual machine by first shutting down Linux and
    then log it off.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'SOFTOFF'
          userid      - userid of the virtual machine
          parms['maxQueries']     - Maximum number of queries to issue.
                                    Optional.
          parms['maxWait']        - Maximum time to wait in seconds.
                                    Optional,
                                    unless 'maxQueries' is specified.
          parms['poll']           - Polling interval in seconds. Optional,
                                    unless 'maxQueries' is specified.

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.softDeactivate, userid: " +
        rh.userid)

    strCmd = "echo 'ping'"
    iucvResults = execCmdThruIUCV(rh, rh.userid, strCmd)

    if iucvResults['overallRC'] == 0:
        # We could talk to the machine, tell it to shutdown nicely.
        strCmd = "shutdown -h now"
        iucvResults = execCmdThruIUCV(rh, rh.userid, strCmd)
        if iucvResults['overallRC'] == 0:
            time.sleep(15)
        else:
            # Shutdown failed.  Let CP take down the system
            # after we log the results.
            rh.printSysLog("powerVM.softDeactivate " + rh.userid +
                " is unreachable. Treating it as already shutdown.")
    else:
        # Could not ping the machine.  Treat it as a success
        # after we log the results.
        rh.printSysLog("powerVM.softDeactivate " + rh.userid +
            " is unreachable. Treating it as already shutdown.")

    # Tell z/VM to log off the system.
    parms = ["-T", rh.userid]
    smcliResults = invokeSMCLI(rh, "Image_Deactivate", parms)
    if smcliResults['overallRC'] == 0:
        pass
    elif (smcliResults['overallRC'] == 8 and smcliResults['rc'] == 200 and
        (smcliResults['rs'] == 12 or + smcliResults['rs'] == 16)):
        # Tolerable error.
        # Machine is already logged off or is logging off.
        rh.printLn("N", rh.userid + " is already logged off.")
    else:
        # SMAPI API failed.
        rh.printLn("ES", smcliResults['response'])
        rh.updateResults(smcliResults)    # Use results from invokeSMCLI

    if rh.results['overallRC'] == 0 and 'maxQueries' in rh.parms:
        # Wait for the system to log off.
        waitResults = waitForVMState(
            rh,
            rh.userid,
            'off',
            maxQueries=rh.parms['maxQueries'],
            sleepSecs=rh.parms['poll'])
        if waitResults['overallRC'] == 0:
            rh.printLn("N", "Userid '" + rh.userid +
                " is in the desired state: off")
        else:
            rh.updateResults(waitResults)

    rh.printSysLog("Exit powerVM.softDeactivate, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def unpause(rh):
    """
    Unpause a virtual machine.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'UNPAUSE'
          userid      - userid of the virtual machine

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.unpause, userid: " + rh.userid)

    parms = ["-T", rh.userid, "-k", "PAUSE=NO"]

    results = invokeSMCLI(rh, "Image_Pause", parms)
    if results['overallRC'] != 0:
        # SMAPI API failed.
        rh.printLn("ES", results['response'])
        rh.updateResults(results)    # Use results from invokeSMCLI

    rh.printSysLog("Exit powerVM.unpause, rc: " +
        str(rh.results['overallRC']))
    return rh.results['overallRC']


def wait(rh):
    """
    Wait for the virtual machine to go into the specified state.

    Input:
       Request Handle with the following properties:
          function    - 'POWERVM'
          subfunction - 'WAIT'
          userid      - userid of the virtual machine
          parms['desiredState']   - Desired state
          parms['maxQueries']     - Maximum number of queries to issue
          parms['maxWait']        - Maximum time to wait in seconds
          parms['poll']           - Polling interval in seconds

    Output:
       Request Handle updated with the results.
       Return code - 0: ok, non-zero: error
    """

    rh.printSysLog("Enter powerVM.wait, userid: " + rh.userid)

    if (rh.parms['desiredState'] == 'off' or
        rh.parms['desiredState'] == 'on'):
        results = waitForVMState(
            rh,
            rh.userid,
            rh.parms['desiredState'],
            maxQueries=rh.parms['maxQueries'],
            sleepSecs=rh.parms['poll'])
    else:
        results = waitForOSState(
            rh,
            rh.userid,
            rh.parms['desiredState'],
            maxQueries=rh.parms['maxQueries'],
            sleepSecs=rh.parms['poll'])

    if results['overallRC'] == 0:
        rh.printLn("N", rh.userid + ": " + rh.parms['desiredState'])
    else:
        rh.updateResults(results)

    rh.printSysLog("Exit powerVM.wait, rc: " + str(rh.results['overallRC']))
    return rh.results['overallRC']



# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import sys


__all__ = ['__version__']
__version__ = '1.0.1'


# Check supported Python versions
_PYTHON_M = sys.version_info[0]
_PYTHON_N = sys.version_info[1]
if _PYTHON_M == 2 and _PYTHON_N < 7:
    raise RuntimeError('On Python 2, zvm sdk requires Python 2.7')
elif _PYTHON_M == 3 and _PYTHON_N > 5:
    raise RuntimeError('On Python 3, zvm sdk supports to Python 3.5')



# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc
import re
import six

from zvmsdk import config
from zvmsdk import exception
from zvmsdk import log
from zvmsdk import vmops


_VolumeOP = None
CONF = config.CONF
LOG = log.LOG

# instance parameters:
NAME = 'name'
OS_TYPE = 'os_type'
# volume parameters:
SIZE = 'size'
TYPE = 'type'
LUN = 'lun'
# connection_info parameters:
ALIAS = 'alias'
PROTOCOL = 'protocol'
FCPS = 'fcps'
WWPNS = 'wwpns'
DEDICATE = 'dedicate'


def get_volumeop():
    global _VolumeOP
    if not _VolumeOP:
        _VolumeOP = VolumeOperator()
    return _VolumeOP


@six.add_metaclass(abc.ABCMeta)
class VolumeOperatorAPI(object):
    """Volume operation APIs oriented towards SDK driver.

    The reason to design these APIs is to facilitate the SDK driver
    issuing a volume related request without knowing details. The
    details among different distributions, different instance status,
    different volume types and so on are all hidden behind these APIs.
    The only thing the issuer need to know is what it want to do on
    which targets.

    In fact, that's an ideal case. In the real world, something like
    connection_info still depends on different complex and the issuer
    needs to know how to deal with its case. Even so, these APIs can
    still make things much easier.
    """

    @abc.abstractmethod
    def attach_volume_to_instance(self, instance, volume, connection_info,
                                  is_rollback_in_failure=False):
        raise NotImplementedError

    @abc.abstractmethod
    def detach_volume_from_instance(self, instance, volume, connection_info,
                                    is_rollback_in_failure=False):
        raise NotImplementedError


@six.add_metaclass(abc.ABCMeta)
class VolumeConfiguratorAPI(object):
    """Volume configure APIs to implement volume config jobs on the
    target instance, like: attach, detach, and so on.

    The reason to design these APIs is to hide the details among
    different Linux distributions and releases.
    """

    @abc.abstractmethod
    def config_attach(self, instance, volume, connection_info):
        raise NotImplementedError

    @abc.abstractmethod
    def config_force_attach(self, instance, volume, connection_info):
        # roll back when detaching fails
        raise NotImplementedError

    @abc.abstractmethod
    def config_detach(self, instance, volume, connection_info):
        raise NotImplementedError

    @abc.abstractmethod
    def config_force_detach(self, instance, volume, connection_info):
        # roll back when attaching fails
        raise NotImplementedError


class _BaseConfigurator(VolumeConfiguratorAPI):
    """Contain common code for all distros."""

    def __init__(self):
        self._vmop = vmops.get_vmops()

    def config_attach(self, instance, volume, connection_info):
        if self._vmop.is_reachable(instance[NAME]):
            self.config_attach_active(instance, volume, connection_info)
        else:
            self.config_attach_inactive(instance, volume, connection_info)

    def config_force_attach(self, instance, volume, connection_info):
        pass

    def config_detach(self, instance, volume, connection_info):
        if self._vmop.is_reachable(instance[NAME]):
            self.config_detach_active(instance, volume, connection_info)
        else:
            self.config_detach_inactive(instance, volume, connection_info)

    def config_force_detach(self, instance, volume, connection_info):
        pass

    def config_attach_active(self, instance, volume, connection_info):
        raise NotImplementedError

    def config_attach_inactive(self, instance, volume, connection_info):
        raise NotImplementedError

    def config_detach_active(self, instance, volume, connection_info):
        raise NotImplementedError

    def config_detach_inactive(self, instance, volume, connection_info):
        raise NotImplementedError


class _Configurator_RHEL7(_BaseConfigurator):

    def config_attach_active(self, instance, volume, connection_info):
        pass

    def config_attach_inactive(self, instance, volume, connection_info):
        pass

    def config_detach_active(self, instance, volume, connection_info):
        pass

    def config_detach_inactive(self, instance, volume, connection_info):
        pass


class _Configurator_SLES12(_BaseConfigurator):

    def config_attach_active(self, instance, volume, connection_info):
        pass

    def config_attach_inactive(self, instance, volume, connection_info):
        pass

    def config_detach_active(self, instance, volume, connection_info):
        pass

    def config_detach_inactive(self, instance, volume, connection_info):
        pass


class _Configurator_Ubuntu16(_BaseConfigurator):

    def config_attach_active(self, instance, volume, connection_info):
        pass

    def config_attach_inactive(self, instance, volume, connection_info):
        pass


class VolumeOperator(VolumeOperatorAPI):

    __singleton = None

    _PATTERN_RHEL7 = '^rhel7(\.[0-9])?$'
    _PATTERN_SLES12 = '^sles12(sp[1-9])?$'
    _PATTERN_UBUNTU16 = '^ubuntu16(\.[0-9][0-9])?$'

    def __new__(cls, *args, **kwargs):
        if not cls.__singleton:
            cls.__singleton = super(VolumeOperator, cls
                                    ).__new__(cls, *args, **kwargs)
        return cls.__singleton

    def attach_volume_to_instance(self,
                                  instance,
                                  volume,
                                  connection_info,
                                  is_rollback_in_failure=False):
        LOG.debug("Enter VolumeOperator.attach_volume_to_instance, attach "
                  "volume %(vol)s to instance %(inst)s by connection_info "
                  "%(conn_info)s.")

        self._validate_instance(instance)
        self._validate_volume(volume)
        self._validate_connection_info(connection_info)

        configurator = self._get_configurator(instance)
        configurator.config_attach(instance, volume, connection_info)

        LOG.debug("Exit VolumeOperator.attach_volume_to_instance.")

    def detach_volume_from_instance(self,
                                    instance,
                                    volume,
                                    connection_info,
                                    is_rollback_in_failure=False):
        LOG.debug("Enter VolumeOperator.detach_volume_from_instance, detach "
                  "volume %(vol)s from instance %(inst)s by connection_info "
                  "%(conn_info)s.")

        self._validate_instance(instance)
        self._validate_volume(volume)
        self._validate_connection_info(connection_info)

        configurator = self._get_configurator(instance)
        configurator.config_detach(instance, volume, connection_info)

        LOG.debug("Exit VolumeOperator.detach_volume_from_instance.")

    def _validate_instance(self, instance):
        # arg type checks are done in API level

        if OS_TYPE not in instance.keys():
            raise exception.SDKInvalidInputFormat(
                "instance os_type is not passed in!")

        if NAME not in instance.keys():
            raise exception.SDKInvalidInputFormat(
                "instance name is not passed in!")

        # os_type patterns for rhel7, sles12 and ubuntu 16, i.e.
        # rhel7: rhel7, rhel7.2
        # sles12: sles12, sles12sp2
        # ubuntu16: ubuntu16, ubuntu16.04
        os_type_pattern = (self._PATTERN_RHEL7 +
                           '|' + self._PATTERN_SLES12 +
                           '|' + self._PATTERN_UBUNTU16)
        if not re.match(os_type_pattern, instance[OS_TYPE]):
            msg = ("unknown instance os_type: %s . It must be one of set "
                   "'rhel7, sles12, or ubuntu16'." % instance[OS_TYPE])
            raise exception.SDKInvalidInputFormat(msg)

    def _is_16bit_hex(self, value):
        pattern = '^[0-9a-f]{16}$'
        try:
            return re.match(pattern, value)
        except Exception:
            return False

    def _validate_volume(self, volume):
        # arg type checks are done in API level

        if TYPE not in volume.keys():
            raise exception.SDKInvalidInputFormat(
                "volume type is not passed in!")

        # support only FiberChannel volumes at this moment
        if volume[TYPE] != 'fc':
            msg = ("volume type: %s is illegal!" % volume[TYPE])
            raise exception.SDKInvalidInputFormat(msg)
        self._validate_fc_volume(volume)

    def _validate_fc_volume(self, volume):
        # exclusively entering from _validate_volume at this moment, so will
        # not check volume object again. Modify it if necessary in the future
        if LUN not in volume.keys():
            raise exception.SDKInvalidInputFormat(
                "volume LUN is not passed in!")

        volume[LUN] = volume[LUN].lower()
        if not self._is_16bit_hex(volume[LUN]):
            msg = ("volume LUN value: %s is illegal!" % volume[LUN])
            raise exception.SDKInvalidInputFormat(msg)

    def _validate_connection_info(self, connection_info):
        # arg type checks are done in API level

        if ALIAS not in connection_info.keys():
            raise exception.SDKInvalidInputFormat(
                "device alias is not passed in!")

        if PROTOCOL not in connection_info.keys():
            raise exception.SDKInvalidInputFormat(
                "connection protocol is not passed in!")

        # support only FiberChannel volumes at this moment
        if connection_info[PROTOCOL] != 'fc':
            raise exception.SDKInvalidInputFormat(
                "connection protocol: %s is illegal!" %
                connection_info[PROTOCOL])
        self._validate_fc_connection_info(connection_info)

    def _validate_fc_connection_info(self, connection_info):
        # exclusively entering from _validate_connection_info at this moment,
        # so will not check connection_info object again. Modify it if
        # necessary in the future
        if DEDICATE in connection_info.keys():
            if not isinstance(connection_info[DEDICATE], list):
                msg = ("dedicate devices in connection info must be of type "
                       "list, however the object passed in is: %s !"
                       % connection_info[DEDICATE])
                raise exception.SDKInvalidInputFormat(msg)
            for i in range(len(connection_info[DEDICATE])):
                connection_info[DEDICATE][i] = (
                        connection_info[DEDICATE][i].lower())

        if FCPS not in connection_info.keys():
            raise exception.SDKInvalidInputFormat(
                "fcp devices are not passed in!")

        if not isinstance(connection_info[FCPS], list):
            msg = ("fcp devices in connection info must be of type list, "
                   "however the object passed in is: %s !"
                   % connection_info[FCPS])
            raise exception.SDKInvalidInputFormat(msg)

        if WWPNS not in connection_info.keys():
            raise exception.SDKInvalidInputFormat("WWPNS are not passed in!")

        if not isinstance(connection_info[WWPNS], list):
            msg = ("wwpns in connection info must be of type list, however "
                   "the object passed in is: %s !" % connection_info[WWPNS])
            raise exception.SDKInvalidInputFormat(msg)

        for i in range(len(connection_info[FCPS])):
            connection_info[FCPS][i] = connection_info[FCPS][i].lower()
        for fcp in connection_info[FCPS]:
            self._validate_fcp(fcp)

        for i in range(len(connection_info[WWPNS])):
            connection_info[WWPNS][i] = connection_info[WWPNS][i].lower()
        for wwpn in connection_info[WWPNS]:
            if not self._is_16bit_hex(wwpn):
                msg = ("WWPN value: %s is illegal!" % wwpn)
                raise exception.SDKInvalidInputFormat(msg)

    def _validate_fcp(self, fcp):
        # exclusively entering from _validate_fc_connection_info at this
        # moment, so will not check fcp object again. Modify it if necessary
        # in the future
        pattern = '^[0-9a-f]{1,4}$'
        try:
            if not re.match(pattern, fcp):
                raise exception.SDKInvalidInputFormat(
                    "fcp value: %s is illegal!" % fcp)
        except Exception:
            msg = ("fcp object must be of type string, "
                   "however the object passed in is: %s !" % fcp)
            raise exception.SDKInvalidInputFormat(msg)

    def _get_configurator(self, instance):
        # all input object should have been validated by _validate_xxx method
        # before being passed in
        if re.match(self._PATTERN_RHEL7, instance[OS_TYPE]):
            return _Configurator_RHEL7()
        elif re.match(self._PATTERN_SLES12, instance[OS_TYPE]):
            return _Configurator_SLES12()
        elif re.match(self._PATTERN_UBUNTU16, instance[OS_TYPE]):
            return _Configurator_Ubuntu16()
        else:
            raise exception.SDKInvalidInputFormat(
                "unknown instance os: %s!" % instance[OS_TYPE])
